@@ -1,13 +1,15 @@
+from __future__ import annotations
+
 import json
 import uuid
-import logging
 import functools
-import typing
+import typing as tp
 import logging
+from logging import Logger
 from pathlib import Path
 
-from google.cloud.bigquery import dataset
 # hidden BQ and pandas imports due to https://github.com/allegro/bigflow/issues/149
+from typing import Dict, List
 
 
 logger = logging.getLogger(__name__)
@@ -38,7 +40,7 @@ def handle_key_error(method):
     return decorated
 
 
-def get_partition_from_run_datetime_or_none(run_datetime):
+def get_partition_from_run_datetime_or_none(run_datetime: str | None) -> str | None:
     """
     :param run_datetime: string run datetime in format YYYY-MM-DD HH:mm:ss or YYY-MM-DD
     :return: string partition in format YYYYMMDD
@@ -99,7 +101,7 @@ class TemplatedDatasetManager(object):
         return self.dataset_manager.collect(sql.format(**self.template_variables(custom_run_datetime)))
 
     @handle_key_error
-    def collect_list(self, sql: str, custom_run_datetime: typing.Optional[str] = None, record_as_dict: bool = False):
+    def collect_list(self, sql: str, custom_run_datetime: tp.Optional[str] = None, record_as_dict: bool = False):
         return self.dataset_manager.collect_list(
             sql.format(**self.template_variables(custom_run_datetime)), record_as_dict)
 
@@ -133,7 +135,7 @@ class TemplatedDatasetManager(object):
     def create_table_from_schema(
             self,
             table_name: str,
-            schema: typing.Union[None, typing.List[dict], Path] = None,
+            schema: tp.Union[None, tp.List[dict], Path] = None,
             table=None):
         table_id = self.create_table_id(table_name)
         return self.dataset_manager.create_table_from_schema(table_id, schema, table)
@@ -141,7 +143,7 @@ class TemplatedDatasetManager(object):
     def insert(
             self,
             table_name: str,
-            records: typing.Union[typing.List[dict], Path]):
+            records: tp.Union[tp.List[dict], Path]):
         table_id = self.create_table_id(table_name)
         return self.dataset_manager.insert(table_id, records)
 
@@ -182,7 +184,7 @@ class PartitionedDatasetManager(object):
     def collect(self, sql, custom_run_datetime=None):
         return self._dataset_manager.collect(sql, custom_run_datetime)
 
-    def collect_list(self, sql: str, custom_run_datetime: typing.Optional[str] = None, record_as_dict: bool = False):
+    def collect_list(self, sql: str, custom_run_datetime: tp.Optional[str] = None, record_as_dict: bool = False):
         return self._dataset_manager.collect_list(sql, custom_run_datetime, record_as_dict)
 
     def dry_run(self, sql, custom_run_datetime=None):
@@ -229,16 +231,16 @@ class PartitionedDatasetManager(object):
     def create_table_from_schema(
             self,
             table_name: str,
-            schema: typing.Optional[typing.Union[typing.List[dict], Path]] = None,
+            schema: tp.Optional[tp.Union[tp.List[dict], Path]] = None,
             table=None):
         return self._dataset_manager.create_table_from_schema(table_name, schema, table)
 
     def insert(
             self,
             table_name: str,
-            records: typing.Union[typing.List[dict], Path],
+            records: tp.Union[tp.List[dict], Path],
             partitioned: bool = True,
-            custom_run_datetime: typing.Optional[str] = None):
+            custom_run_datetime: tp.Optional[str] = None):
         table_id = self._create_table_id(custom_run_datetime, table_name, partitioned)
         return self._dataset_manager.insert(table_id, records)
 
@@ -261,19 +263,19 @@ class DatasetManager(object):
     Manages BigQuery IO operations.
     """
     def __init__(self,
-                 bigquery_client,
-                 dataset,
-                 logger):
+                 bigquery_client: 'google.cloud.bigquery.Client',
+                 dataset: 'google.cloud.bigquery.Dataset',
+                 logger: logging.Logger):
         from google.cloud import bigquery
-        self.bigquery_client: bigquery.Client  = bigquery_client
+        self.bigquery_client: bigquery.Client = bigquery_client
         self.dataset = dataset
         self.dataset_id = dataset.full_dataset_id.replace(':', '.')
         self.logger = logger
 
-    def write_tmp(self, table_id, sql):
+    def write_tmp(self, table_id: str, sql: str) -> 'google.cloud.bigquery.table.RowIterator':
         return self.write(table_id, sql, 'WRITE_TRUNCATE')
 
-    def write(self, table_id, sql, mode):
+    def write(self, table_id: str, sql: str, mode: str) -> 'google.cloud.bigquery.table.RowIterator':
         from google.cloud import bigquery
         self.logger.info('%s to %s', mode, table_id)
         job_config = bigquery.QueryJobConfig()
@@ -286,20 +288,20 @@ class DatasetManager(object):
         job = self.bigquery_client.query(sql, job_config=job_config)
         return job.result()
 
-    def write_truncate(self, table_id, sql):
+    def write_truncate(self, table_id: str, sql: str) -> 'google.cloud.bigquery.table.RowIterator':
         self.table_exists_or_error(table_id)
         return self.write(table_id, sql, 'WRITE_TRUNCATE')
 
-    def write_append(self, table_id, sql):
+    def write_append(self, table_id: str, sql: str) -> 'google.cloud.bigquery.table.RowIterator':
         self.table_exists_or_error(table_id)
         return self.write(table_id, sql, 'WRITE_APPEND')
 
-    def table_exists_or_error(self, table_id):
+    def table_exists_or_error(self, table_id: str):
         table_name = table_id.split('$')[0].split('.')[2]
         if not self.table_exists(table_name):
             raise ValueError('Table {id} does not exist'.format(id=table_id))
 
-    def create_table(self, create_query):
+    def create_table(self, create_query: str) -> 'google.cloud.bigquery.table.RowIterator':
         from google.cloud import bigquery
         self.logger.info('CREATE TABLE: %s', create_query)
         job_config = bigquery.QueryJobConfig()
@@ -311,16 +313,20 @@ class DatasetManager(object):
             job_config=job_config)
         return job.result()
 
-    def collect(self, sql):
+    def collect(self, sql: str) -> 'pandas.DataFrame':
         return self._query(sql).to_dataframe()
 
-    def collect_list(self, sql: str, record_as_dict: bool = False):
+    def collect_list(
+            self,
+            sql: str,
+            record_as_dict: bool = False
+    ) -> tp.List[tp.Dict] | tp.List['google.cloud.bigquery.table.Row']:
         result = list(self._query(sql).result())
         if record_as_dict:
             result = [dict(e) for e in result]
         return result
 
-    def dry_run(self, sql):
+    def dry_run(self, sql: str) -> str:
         from google.cloud import bigquery
         job_config = bigquery.QueryJobConfig()
         job_config.dry_run = True
@@ -333,10 +339,10 @@ class DatasetManager(object):
     def remove_dataset(self):
         return self.bigquery_client.delete_dataset(self.dataset, delete_contents=True, not_found_ok=True)
 
-    def load_table_from_dataframe(self, table_id, df):
+    def load_table_from_dataframe(self, table_id: str, df) -> 'google.cloud.bigquery.table.RowIterator':
         return self.bigquery_client.load_table_from_dataframe(df, table_id).result()
 
-    def table_exists(self, table_name):
+    def table_exists(self, table_name: str) -> bool:
         return self.bigquery_client.query('''
             SELECT count(*) as table_exists
             FROM `{dataset_id}.__TABLES__`
@@ -351,7 +357,7 @@ class DatasetManager(object):
     def create_table_from_schema(
             self,
             table_id: str,
-            schema: typing.Union[typing.List[dict], Path, None] = None,
+            schema: tp.Union[tp.List[dict], Path, None] = None,
             table=None):
         from google.cloud.bigquery import Table, TimePartitioning
 
@@ -375,7 +381,7 @@ class DatasetManager(object):
     def insert(
             self,
             table_id: str,
-            records: typing.Union[typing.List[dict], Path]):
+            records: tp.Union[tp.List[dict], Path]):
         self.logger.info('INSERTING RECORDS TO TABLE: %s', table_id)
         table = self.bigquery_client.get_table(table_id)
         if isinstance(records, Path):
@@ -385,7 +391,7 @@ class DatasetManager(object):
         if errors:
             raise ValueError(errors)
 
-    def _query(self, sql, job_config=None):
+    def _query(self, sql: str, job_config=None) -> 'google.cloud.bigquery.job.QueryJob':
         self.logger.info('COLLECTING DATA: %s', sql)
         if job_config:
             return self.bigquery_client.query(sql, job_config=job_config)
@@ -408,20 +414,40 @@ class DatasetManager(object):
                 'cost': cost}
 
 
-def create_dataset(dataset_name, bigquery_client, location=DEFAULT_LOCATION):
+def create_dataset(
+        dataset_name: str,
+        bigquery_client: 'google.cloud.bigquery.Client',
+        location: str = DEFAULT_LOCATION,
+        dataset_new_labels: Dict[str, str] | None = None
+) -> 'google.cloud.bigquery.Dataset':
     from google.cloud import bigquery
     dataset = bigquery.Dataset('{project_id}.{dataset_name}'.format(
         project_id=bigquery_client.project,
         dataset_name=dataset_name))
     dataset.location = location
-    return bigquery_client.create_dataset(dataset, exists_ok=True)
+    bigquery_dataset = bigquery_client.create_dataset(dataset, exists_ok=True)
+    if dataset_new_labels:
+        logger.info(f'ADDING LABELS FOR DATASET: ${dataset_name}')
+        bigquery_dataset.labels = _prepare_labels(bigquery_dataset.labels, dataset_new_labels)
+        return bigquery_client.update_dataset(bigquery_dataset, ["labels"])
+    return bigquery_dataset
 
 
-def random_uuid(suffix=''):
+def _prepare_labels(labels: Dict[str, str], new_labels: Dict[str, str]) -> Dict[str, str]:
+    labels_to_remove = {label_to_remove: None for label_to_remove in labels}
+    labels_to_remove.update(new_labels)
+    return labels_to_remove
+
+
+def random_uuid(suffix='') -> str:
     return uuid.uuid1().hex + suffix
 
 
-def create_bigquery_client(project_id, credentials, location):
+def create_bigquery_client(
+        project_id: str,
+        credentials: 'google.auth.credentials.Credentials' | None,
+        location: str
+) -> 'google.cloud.bigquery.Client':
     from google.cloud import bigquery
     return bigquery.Client(
         project=project_id,
@@ -429,16 +455,35 @@ def create_bigquery_client(project_id, credentials, location):
         location=location)
 
 
+def upsert_tables_labels(
+        dataset_name: str,
+        tables_labels: Dict[str, Dict[str, str]] | None,
+        bigquery_client: 'google.cloud.bigquery.Client'):
+    from google.cloud.exceptions import NotFound
+    if tables_labels:
+        for table_name, labels in tables_labels.items():
+            logger.info(f'ADDING LABELS FOR TABLE: ${table_name}')
+            try:
+                table = bigquery_client.get_table(dataset_name + '.' + table_name)
+            except NotFound:
+                continue
+            table.labels = _prepare_labels(table.labels, tables_labels[table_name])
+            bigquery_client.update_table(table, ["labels"])
+
+
 def create_dataset_manager(
-        project_id,
-        runtime,
-        dataset_name=None,
-        internal_tables=None,
-        external_tables=None,
-        extras=None,
-        credentials=None,
-        location=DEFAULT_LOCATION,
-        logger=None) -> typing.Tuple[str, PartitionedDatasetManager]:
+        project_id: str,
+        runtime: str,
+        dataset_name: str | None = None,
+        internal_tables: List[str] | None = None,
+        external_tables: Dict[str, str] | None = None,
+        extras: Dict | None = None,
+        credentials: 'google.auth.credentials.Credentials' | None = None,
+        location: str = DEFAULT_LOCATION,
+        logger: Logger | None = None,
+        tables_labels: Dict[str, Dict[str, str]] | None = None,
+        dataset_labels: Dict[str, str] | None = None
+) -> tp.Tuple[str, PartitionedDatasetManager]:
     """
     Dataset manager factory.
     If dataset does not exist then it will also create dataset with given name.
@@ -454,6 +499,8 @@ def create_dataset_manager(
     :param project_id: string full project id where dataset being processed is available.
     :param location: string location of dataset will be used to create datasets, tables, jobs, etc. EU by default.
     :param logger: custom logger.
+    :param tables_labels: Dict with key as table_name and value as list of key/valued labels.
+    :param dataset_labels: Dict with key/valued labels.
     :return: tuple (full dataset ID, dataset manager).
     """
     dataset_name = dataset_name or random_uuid(suffix='_test_case')
@@ -465,7 +512,9 @@ def create_dataset_manager(
         logger = logging.getLogger(__name__)
 
     client = create_bigquery_client(project_id, credentials, location)
-    dataset = create_dataset(dataset_name, client, location)
+    dataset = create_dataset(dataset_name, client, location, dataset_labels)
+
+    upsert_tables_labels(dataset_name, tables_labels, client)
 
     core_dataset_manager = DatasetManager(client, dataset, logger)
     templated_dataset_manager = TemplatedDatasetManager(core_dataset_manager, internal_tables, external_tables, extras, runtime)
