@@ -1,10 +1,13 @@
-from unittest import TestCase, mock
-import itertools
+from unittest import mock
 import shutil
 import freezegun
 
+from bigflow.build.operate import BuildImageCacheParams
+from bigflow.deploy import AuthorizationType
+
 from bigflow.testing.isolate import ForkIsolateMixin
 from bigflow.cli import *
+from bigflow.cli import _ConsoleStreamLogHandler
 
 from test import mixins
 
@@ -14,7 +17,7 @@ TESTS_DIR = Path(__file__).parent
 class CliTestCase(
     mixins.PrototypedDirMixin,
     ForkIsolateMixin,
-    TestCase,
+    mixins.BaseTestCase,
 ):
     proto_dir = "bf-projects/example_project"
 
@@ -268,7 +271,7 @@ class CliTestCase(
             # when
             cli(['deploy-dags'])
 
-    @mock.patch('bigflow.cli.deploy_dags_folder')
+    @mock.patch('bigflow.deploy.deploy_dags_folder')
     def test_should_call_cli_deploy_dags_command__with_defaults_and_with_implicit_deployment_config_file(self,
                                                                                                          deploy_dags_folder_mock):
         # given
@@ -296,7 +299,7 @@ deployment_config = Config(name='dev',
                                                    vault_endpoint=None,
                                                    vault_secret='secret')
 
-    @mock.patch('bigflow.cli.deploy_dags_folder')
+    @mock.patch('bigflow.deploy.deploy_dags_folder')
     def test_should_call_cli_deploy_dags_command_for_different_environments(self, deploy_dags_folder_mock):
         # given
         self._touch_file('deployment_config.py',
@@ -354,7 +357,7 @@ deployment_config = Config(name='dev',
                                                    vault_endpoint=None,
                                                    vault_secret='secret-prod')
 
-    @mock.patch('bigflow.cli.deploy_dags_folder')
+    @mock.patch('bigflow.deploy.deploy_dags_folder')
     def test_should_call_cli_deploy_dags_command__when_parameters_are_given_by_explicit_deployment_config_file(self,
                                                                                                                deploy_dags_folder_mock):
         # given
@@ -387,7 +390,7 @@ deployment_config = Config(name='dev',
                                                    vault_endpoint='my-another-vault-endpoint',
                                                    vault_secret='secrett')
 
-    @mock.patch('bigflow.cli.deploy_dags_folder')
+    @mock.patch('bigflow.deploy.deploy_dags_folder')
     def test_should_call_cli_deploy_dags_command__when_all_parameters_are_given_by_cli_arguments(self,
                                                                                                  deploy_dags_folder_mock):
         # when
@@ -410,7 +413,7 @@ deployment_config = Config(name='dev',
                                                    vault_endpoint='my-vault-endpoint',
                                                    vault_secret='secrett')
 
-    @mock.patch('bigflow.cli.deploy_docker_image')
+    @mock.patch('bigflow.deploy.deploy_docker_image')
     def test_should_call_cli_deploy_image_command__with_defaults_and_with_implicit_deployment_config_file(self,
                                                                                                           deploy_docker_image_mock):
         # given
@@ -434,7 +437,7 @@ deployment_config = Config(name='dev',
                                                     vault_endpoint=None,
                                                     vault_secret=None)
 
-    @mock.patch('bigflow.cli.deploy_docker_image')
+    @mock.patch('bigflow.deploy.deploy_docker_image')
     def test_should_call_cli_deploy_image_command__with_explicit_deployment_config_file(self, deploy_docker_image_mock):
         # given
         dc_file = self._touch_file('my_deployment_config.py',
@@ -463,7 +466,7 @@ deployment_config = Config(name='dev',
                                                     vault_endpoint='my-another-vault-endpoint',
                                                     vault_secret='secrett')
 
-    @mock.patch('bigflow.cli.deploy_docker_image')
+    @mock.patch('bigflow.deploy.deploy_docker_image')
     def test_should_call_cli_deploy_image_command__when_all_parameters_are_given_by_cli_arguments_and_image_is_loaded_from_tar(
             self, deploy_docker_image_mock):
         # when
@@ -482,7 +485,7 @@ deployment_config = Config(name='dev',
                                                     vault_endpoint='my-vault-endpoint',
                                                     vault_secret='secrett')
 
-    @mock.patch('bigflow.cli.deploy_docker_image')
+    @mock.patch('bigflow.deploy.deploy_docker_image')
     def test_should_find_tar_in_image_directory(self, deploy_docker_image_mock):
         # given
 
@@ -504,8 +507,32 @@ deployment_config = Config(name='dev',
                                                     vault_endpoint='my-vault-endpoint',
                                                     vault_secret='secrett')
 
-    @mock.patch('bigflow.cli.deploy_dags_folder')
-    @mock.patch('bigflow.cli.deploy_docker_image')
+    @mock.patch('bigflow.deploy.deploy_docker_image')
+    def test_should_find_toml_ref_in_image_directory(self, deploy_docker_image_mock):
+
+        # given
+        shutil.rmtree(Path.cwd() / ".image", ignore_errors=True)
+        self._touch_file('imageinfo-123.toml', '', '.image')
+
+        # when
+        cli(['deploy-image',
+             '--docker-repository', 'my-docker-repository',
+             '--vault-endpoint', 'my-vault-endpoint',
+             '--auth-method', 'vault',
+             '--vault-secret', 'secrett',
+             ])
+
+        # then
+        deploy_docker_image_mock.assert_called_with(
+            auth_method=AuthorizationType.VAULT,
+            docker_repository='my-docker-repository',
+            image_tar_path='.image/imageinfo-123.toml',
+            vault_endpoint='my-vault-endpoint',
+            vault_secret='secrett',
+        )
+
+    @mock.patch('bigflow.deploy.deploy_dags_folder')
+    @mock.patch('bigflow.deploy.deploy_docker_image')
     def test_should_call_both_deploy_methods_with_deploy_command(self, deploy_docker_image_mock,
                                                                  deploy_dags_folder_mock):
         # given
@@ -611,15 +638,110 @@ deployment_config = Config(name='dev',
             read_project_spec.return_value,
             start_time="2001-02-03 15:00:00",
             workflow_id=None,
+            cache_params=None,
+            export_image_tar=None,
         )
 
-    @mock.patch('bigflow.cli._cli_build_image')
-    def test_should_call_cli_build_image_command(self, _cli_build_image_mock):
+    def test_should_call_cli_build_image_command_without_tar(self):
+        # given
+        cli_build_mock = self.addMock(mock.patch('bigflow.cli._cli_build_image'))
+
         # when
-        cli(['build-image'])
+        cli(['build-image', '--no-export-image-tar'])
 
         # then
-        _cli_build_image_mock.assert_called_with(Namespace(operation='build-image', verbose=False))
+        cli_build_mock.assert_called_once()
+        self.assertEqual(cli_build_mock.call_args[0][0].export_image_tar, False)
+
+    def test_should_call_cli_build_image_command_with_tar(self):
+        # given
+        cli_build_mock = self.addMock(mock.patch('bigflow.cli._cli_build_image'))
+
+        # when
+        cli(['build-image', '--export-image-tar'])
+
+        # then
+        cli_build_mock.assert_called_once()
+        self.assertEqual(cli_build_mock.call_args[0][0].export_image_tar, True)
+
+    @mock.patch('bigflow.cli._cli_build_image')
+    def test_should_call_cli_build_image_command_without_tar(self, _cli_build_image_mock):
+        # when
+        cli(['build-image', '--no-export-image-tar'])
+
+        # then
+        _cli_build_image_mock.assert_called_with(
+            Namespace(
+                auth_method=AuthorizationType.LOCAL_ACCOUNT,
+                cache_from_image=None,
+                cache_from_version=None,
+                config=None,
+                deployment_config_path=None,
+                export_image_tar=False,
+                operation='build-image',
+                vault_endpoint=None,
+                vault_secret=None,
+                verbose=False,
+            )
+        )
+
+    def test_should_call_cli_build_image_with_cached_from_image(self):
+
+        # given
+        self.addMock(mock.patch('bigflow.build.spec.read_project_spec'))
+        build_image_mock = self.addMock(mock.patch('bigflow.build.operate.build_image'))
+
+        # when
+        cli([
+            'build-image',
+            '--no-export-image-tar',
+            '--vault-endpoint', 'my-vault-endpoint',
+            '--auth-method', 'vault',
+            '--vault-secret', 'secrett',
+            '--cache-from-image', 'xyz.org/foo:bar',
+            '--cache-from-image', 'xyz.org/foo:baz',
+        ])
+
+        # then
+        build_image_mock.assert_called_once()
+        _, kwrgs = build_image_mock.call_args
+        self.assertEqual(kwrgs['export_image_tar'], False)
+        self.assertEqual(kwrgs['cache_params'], BuildImageCacheParams(
+            auth_method=AuthorizationType.VAULT,
+            vault_endpoint='my-vault-endpoint',
+            vault_secret='secrett',
+            cache_from_image=['xyz.org/foo:bar', 'xyz.org/foo:baz'],
+            cache_from_version=None,
+        ))
+
+    def test_should_call_cli_build_image_with_cached_from_version(self):
+
+        # given
+        self.addMock(mock.patch('bigflow.build.spec.read_project_spec'))
+        build_image_mock = self.addMock(mock.patch('bigflow.build.operate.build_image'))
+
+        # when
+        cli([
+            'build-image',
+            '--no-export-image-tar',
+            '--vault-endpoint', 'my-vault-endpoint',
+            '--auth-method', 'vault',
+            '--vault-secret', 'secrett',
+            '--cache-from-version', 'bar',
+            '--cache-from-version', 'baz',
+        ])
+
+        # then
+        build_image_mock.assert_called_once()
+        _, kwrgs = build_image_mock.call_args
+        self.assertEqual(kwrgs['export_image_tar'], False)
+        self.assertEqual(kwrgs['cache_params'], BuildImageCacheParams(
+            auth_method=AuthorizationType.VAULT,
+            vault_endpoint='my-vault-endpoint',
+            vault_secret='secrett',
+            cache_from_image=None,
+            cache_from_version=['bar', 'baz'],
+        ))
 
     @mock.patch('bigflow.build.operate.build_project')
     @mock.patch('bigflow.build.spec.read_project_spec')
@@ -638,6 +760,8 @@ deployment_config = Config(name='dev',
             read_project_spec.return_value,
             start_time="2001-02-03 15:00:00",
             workflow_id=None,
+            cache_params=None,
+            export_image_tar=None,
         )
 
     @mock.patch('bigflow.cli._cli_build_package')
@@ -652,22 +776,38 @@ deployment_config = Config(name='dev',
     def test_should_call_cli_build_command(self, _cli_build_mock):
         # when
         cli(['build'])
+        args = Namespace(
+            auth_method=AuthorizationType.LOCAL_ACCOUNT,
+            cache_from_image=None,
+            cache_from_version=None,
+            deployment_config_path=None,
+            export_image_tar=None,
+            operation='build',
+            start_time=None,
+            vault_endpoint=None,
+            vault_secret=None,
+            verbose=False,
+            workflow=None,
+            config=None,
+        )
 
         # then
-        _cli_build_mock.assert_called_with(Namespace(operation='build', start_time=None, workflow=None, verbose=False))
+        _cli_build_mock.assert_called_with(args)
 
         # when
         cli(['build', '--start-time', '2020-01-01 00:00:00'])
 
         # then
-
-        _cli_build_mock.assert_called_with(Namespace(operation='build', start_time='2020-01-01 00:00:00', workflow=None, verbose=False))
+        args.start_time = '2020-01-01 00:00:00'
+        _cli_build_mock.assert_called_with(args)
 
         # when
         cli(['build', '--start-time', '2020-01-01 00:00:00', '--workflow', 'some_workflow'])
 
         # then
-        _cli_build_mock.assert_called_with(Namespace(operation='build', start_time='2020-01-01 00:00:00', workflow='some_workflow', verbose=False))
+        args.start_time = '2020-01-01 00:00:00'
+        args.workflow = 'some_workflow'
+        _cli_build_mock.assert_called_with(args)
 
     @mock.patch('bigflow.build.operate.build_package')
     @mock.patch('bigflow.build.spec.read_project_spec')
@@ -696,7 +836,7 @@ deployment_config = Config(name='dev',
 
         # then
         read_project_mock.assert_called_once()
-        build_image_mock.assert_any_call(read_project_mock.return_value)
+        build_image_mock.assert_called_once()
 
     @mock.patch('bigflow.build.operate.build_dags')
     @mock.patch('bigflow.build.spec.read_project_spec')
@@ -717,7 +857,7 @@ deployment_config = Config(name='dev',
             workflow_id=None,
         )
 
-    @mock.patch('bigflow.cli.get_version')
+    @mock.patch('bigflow.version.get_version')
     def test_should_call_cli_project_version_command(self, get_version):
         # when
         cli(['project-version'])
@@ -725,7 +865,7 @@ deployment_config = Config(name='dev',
         # then
         get_version.assert_called_once()
 
-    @mock.patch('bigflow.cli.get_version')
+    @mock.patch('bigflow.version.get_version')
     def test_should_call_cli_project_version_command_by_alias(self, get_version):
         # when
         cli(['pv'])
@@ -733,7 +873,7 @@ deployment_config = Config(name='dev',
         # then
         get_version.assert_called_once()
 
-    @mock.patch('bigflow.cli.release')
+    @mock.patch('bigflow.version.release')
     def test_should_call_cli_release_command_with_no_args(self, release):
         # when
         cli(['release'])
@@ -741,7 +881,7 @@ deployment_config = Config(name='dev',
         # then
         release.assert_called_once_with(None)
 
-    @mock.patch('bigflow.cli.release')
+    @mock.patch('bigflow.version.release')
     def test_should_call_cli_release_command_with_identity_file(self, release):
         # when
         cli(['release', '--ssh-identity-file', 'path/to/identity_file'])
@@ -749,7 +889,7 @@ deployment_config = Config(name='dev',
         # then
         release.assert_called_once_with('path/to/identity_file')
 
-    @mock.patch('bigflow.cli.release')
+    @mock.patch('bigflow.version.release')
     def test_should_call_cli_release_command_with_identity_file_parameter_shortcut(self, release):
         # when
         cli(['release', '-i', 'path/to/identity_file'])
@@ -814,74 +954,6 @@ another-project-id                         ANOTHER PROJECT                002242
         # then
         self.assertEqual(cli_start_project_mock.call_count, 1)
 
-    @mock.patch('builtins.print')
-    def test_should_call_cli_logs_command(self, print_mock):
-        # when
-        root_package = TESTS_DIR / "cli_logs_regular_workflows"
-        cli_logs(root_package)
-
-        # then
-        print_mock.assert_called_with(
-            '\n*************************LOGS LINK*************************\nInfrastructure logs:'
-            '\nsome-project-id: https://console.cloud.google.com/logs/query;query=%28'
-            'severity%3E%3D%22INFO%22%0A'
-            'resource.type%3D%22k8s_pod%22%0A%22'
-            'Error%3A%22%29%0AOR%0A%28'
-            'resource.type%3D%22dataflow_step%22%0A'
-            'log_name%3D%22projects%2Fsome-project-id%2Flogs%2Fdataflow.googleapis.com%252Fjob-message%22%0A'
-            'severity%3E%3D%22INFO%22%0A%29\n'
-            'another-project-id: https://console.cloud.google.com/logs/query;query=%28'
-            'severity%3E%3D%22INFO%22%0A'
-            'resource.type%3D%22k8s_pod%22%0A%22'
-            'Error%3A%22%29%0AOR%0A%28'
-            'resource.type%3D%22dataflow_step%22%0A'
-            'log_name%3D%22projects%2Fanother-project-id%2Flogs%2Fdataflow.googleapis.com%252Fjob-message%22%0A'
-            'severity%3E%3D%22INFO%22%0A%29\nWorkflow logs:\nID_1: https://console.cloud.google.com/logs/query;query='
-            'logName%3D%22projects%2Fsome-project-id%2Flogs%2FID_1%22%0Alabels.workflow_id%3D%22ID_1%22%0A'
-            '\nID_2: https://console.cloud.google.com/logs/query;query='
-            'logName%3D%22projects%2Fanother-project-id%2Flogs%2FID_2%22%0Alabels.workflow_id%3D%22ID_2%22%0A'
-            '\n***********************************************************')
-
-    @mock.patch('builtins.print')
-    def test_should_call_cli_logs_and_use_log_name_if_provided(self, print_mock):
-        # when
-        root_package = TESTS_DIR / "cli_logs_log_name_workflow"
-        cli_logs(root_package)
-
-        # then
-        print_mock.assert_called_with(
-            '\n*************************LOGS LINK*************************\nInfrastructure logs:\nsome-project-id: '
-            'https://console.cloud.google.com/logs/query;query=%28'
-            'severity%3E%3D%22INFO%22%0A'
-            'resource.type%3D%22k8s_pod%22%0A%22'
-            'Error%3A%22%29%0AOR%0A%28'
-            'resource.type%3D%22dataflow_step%22%0A'
-            'log_name%3D%22projects%2Fsome-project-id%2Flogs%2Fdataflow.googleapis.com%252Fjob-message%22%0A'
-            'severity%3E%3D%22INFO%22%0A%29\nWorkflow logs:\nID_1: https://console.cloud.google.com/logs/query;query='
-            'logName%3D%22projects%2Fsome-project-id%2Flogs%2Fname-log%22%0A'
-            'labels.workflow_id%3D%22ID_1%22%0A\n***********************************************************')
-
-    @mock.patch('builtins.print')
-    def test_should_deduplicate_projects_id(self, print_mock):
-        # when
-        root_package = TESTS_DIR / "cli_logs_duplicated_workflows"
-        cli_logs(root_package)
-
-        # then
-        print_mock.assert_called_with(
-            '\n*************************LOGS LINK*************************\nInfrastructure logs:'
-            '\nsome-project-id: https://console.cloud.google.com/logs/query;query=%28'
-            'severity%3E%3D%22INFO%22%0A'
-            'resource.type%3D%22k8s_pod%22%0A%22'
-            'Error%3A%22%29%0AOR%0A%28'
-            'resource.type%3D%22dataflow_step%22%0A'
-            'log_name%3D%22projects%2Fsome-project-id%2Flogs%2Fdataflow.googleapis.com%252Fjob-message%22%0A'
-            'severity%3E%3D%22INFO%22%0A%29\nWorkflow logs:\nID_1: https://console.cloud.google.com/logs/query;query='
-            'logName%3D%22projects%2Fsome-project-id%2Flogs%2FID_1%22%0A'
-            'labels.workflow_id%3D%22ID_1%22%0A\nID_2: https://console.cloud.google.com/logs/query;query='
-            'logName%3D%22projects%2Fsome-project-id%2Flogs%2FID_2%22%0Alabels.workflow_id%3D%22ID_2%22%0A'
-            '\n***********************************************************')
-
     def test_should_raise_exception_if_no_workflow_with_log_config_found(self):
         root_package = TESTS_DIR / "i_do_not_exist_at_all"
         with self.assertRaises(Exception) as e:
@@ -924,7 +996,8 @@ another-project-id                         ANOTHER PROJECT                002242
 
         # then
         self.assertEqual(len(logging.root.handlers), 1)
-        self.assertIsInstance(logging.root.handlers[0], logging.StreamHandler)
+
+        self.assertIsInstance(logging.root.handlers[0], _ConsoleStreamLogHandler)
         self.assertEqual(logging.root.level, logging.INFO)
 
     @mock.patch('bigflow.cli._cli_build')
@@ -935,5 +1008,6 @@ another-project-id                         ANOTHER PROJECT                002242
 
         # then
         self.assertEqual(len(logging.root.handlers), 1)
-        self.assertIsInstance(logging.root.handlers[0], logging.StreamHandler)
+        self.assertIsInstance(logging.root.handlers[0], _ConsoleStreamLogHandler)
+
         self.assertEqual(logging.root.level, logging.DEBUG)
